package cc.fascinated.model.user;

import cc.fascinated.common.DateUtils;
import cc.fascinated.model.user.statistic.Statistic;
import cc.fascinated.platform.Platform;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import lombok.*;
import org.springframework.data.annotation.Id;
import org.springframework.data.mongodb.core.index.Indexed;
import org.springframework.data.mongodb.core.mapping.Document;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

/**
 * @author Fascinated (fascinated7)
 */
@RequiredArgsConstructor
@Getter
@Setter
@ToString
@Document("user")
public class User {
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("EEE MMM d HH:mm:ss zzz yyyy");

    /**
     * The ID of the user.
     */
    @Id
    @JsonIgnore
    private final UUID id;

    /**
     * The username of the user.
     * <p>
     * Usually their Steam name.
     * </p>
     */
    @Indexed
    private String username;

    /**
     * The ID of the users steam profile.
     */
    @Indexed
    @JsonProperty("id")
    private String steamId;

    /**
     * Whether the user has logged into the website.
     * <p>
     * This is used to determine if we should track their profiles or not.
     * If they haven't logged in, we don't want to track their profiles.
     * </p>
     */
    @JsonIgnore
    public boolean linkedAccount;

    /**
     * The user's ScoreSaber account.
     */
    public ScoreSaberAccount scoresaberAccount;

    /**
     * The user's statistic history.
     */
    public Map<Platform.Platforms, Map<String, Statistic>> statistics;

    /**
     * The user's statistic history.
     */
    @JsonIgnore
    public Map<Platform.Platforms, Map<Date, Statistic>> getStatistics() {
        if (this.statistics == null) {
            this.statistics = new HashMap<>();
        }
        Map<Platform.Platforms, Map<Date, Statistic>> toReturn = new HashMap<>();
        for (Platform.Platforms platform : statistics.keySet()) {
            toReturn.put(platform, getStatistic(platform));
        }
        return toReturn;
    }

    /**
     * Gets the statistics for a platform.
     *
     * @param platform the platform to get the statistics for
     * @return the statistics
     */
    @SneakyThrows
    public Map<Date, Statistic> getStatistic(@NonNull Platform.Platforms platform) {
        if (this.statistics == null) {
            this.statistics = new HashMap<>();
        }
        Map<String, Statistic> statisticMap = this.statistics.computeIfAbsent(platform, k -> new HashMap<>());
        Map<Date, Statistic> statistics = new HashMap<>();
        for (Map.Entry<String, Statistic> entry : statisticMap.entrySet()) {
            statistics.put(DATE_FORMAT.parse(entry.getKey()), entry.getValue());
        }
        return statistics;
    }

    /**
     * Adds a statistic to the user's history.
     *
     * @param platform the platform to add the statistic for
     * @param date the date of the statistic
     * @param statistic the statistic to add
     */
    public void addStatistic(@NonNull Platform.Platforms platform, @NonNull Date date, @NonNull Statistic statistic) {
        if (this.statistics == null) {
            this.statistics = new HashMap<>();
        }
        Map<String, Statistic> statisticMap = this.statistics.computeIfAbsent(platform, k -> new HashMap<>());
        statisticMap.put(String.valueOf(date.toString()), statistic);
    }

    /**
     * Gets the user as a DTO.
     *
     * @return the user as a DTO
     */
    @JsonIgnore
    public UserDTO getAsDTO() {
        return new UserDTO(this.id, this.username, this.steamId, this.scoresaberAccount);
    }
}
