package cc.fascinated.platform;

import cc.fascinated.exception.impl.BadRequestException;
import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.Setter;

import java.util.Map;

/**
 * @author Fascinated (fascinated7)
 */
@RequiredArgsConstructor
@Getter
@Setter
public abstract class Platform {
    /**
     * The name of the platform.
     */
    private final Platforms platform;

    /**
     * The previous curve version for this
     * platform in the database.
     */
    private int previousCurveVersion;

    /**
     * The current curve version for getting
     * pp from a star count.
     */
    private final int currentCurveVersion;

    /**
     * The curve points for each curve version.
     */
    private final Map<Integer, CurvePoint[]> curvePoints;

    /**
     * Checks the curve version to see if it exists.
     *
     * @param curveVersion the curve version to check
     * @throws BadRequestException if the curve version does not exist
     */
    public void checkCurveVersion(int curveVersion) {
        if (!curvePoints.containsKey(curveVersion)) {
            throw new BadRequestException("Curve version '%s' for platform '%s' was not found.".formatted(curveVersion, platform.getPlatformName()));
        }
    }

    /**
     * Gets the curve points for a curve version.
     *
     * @param curveVersion the curve version to get the curve points for
     * @return the curve points
     */
    public CurvePoint[] getCurve(int curveVersion) {
        this.checkCurveVersion(curveVersion);
        return curvePoints.get(curveVersion);
    }

    /**
     * Gets the PP amount from the star count.
     *
     * @param stars the amount of stars
     * @return the pp amount
     */
    public abstract double getPp(double stars, double accuracy);

    /**
     * Called to update the players
     * data in QuestDB.
     */
    public abstract void updatePlayers();

    /**
     * Called to update the metrics
     * for total scores, etc.
     */
    public abstract void updateMetrics();

    /**
     * Called to update the leaderboards.
     */
    public abstract void updateLeaderboards();

    @AllArgsConstructor
    @Getter
    public enum Platforms {
        SCORESABER("scoresaber");

        /**
         * The internal name of the platform.
         */
        private final String platformName;

        /**
         * Gets a platform by its name.
         *
         * @param platform the name of the platform
         * @return the platform
         */
        public static Platforms getPlatform(String platform) {
            for (Platforms platforms : Platforms.values()) {
                if (platforms.getPlatformName().equalsIgnoreCase(platform)) {
                    return platforms;
                }
            }
            throw new BadRequestException("Platform '" + platform + "' was not found.");
        }
    }
}
