package cc.fascinated.platform.impl;

import cc.fascinated.common.DateUtils;
import cc.fascinated.common.MathUtils;
import cc.fascinated.model.score.Score;
import cc.fascinated.model.token.ScoreSaberAccountToken;
import cc.fascinated.model.token.ScoreSaberLeaderboardPageToken;
import cc.fascinated.model.token.ScoreSaberLeaderboardToken;
import cc.fascinated.model.user.User;
import cc.fascinated.model.user.statistic.impl.ScoreSaberStatistic;
import cc.fascinated.platform.CurvePoint;
import cc.fascinated.platform.Platform;
import cc.fascinated.services.ScoreSaberService;
import cc.fascinated.services.ScoreService;
import cc.fascinated.services.UserService;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.DependsOn;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Fascinated (fascinated7)
 */
@Component
@DependsOn("scoreService")
@Log4j2
public class ScoreSaberPlatform extends Platform {
    /**
     * The base multiplier for stars.
     */
    private final double starMultiplier = 42.114296;

    /**
     * The ScoreSaber service to use
     */
    @NonNull
    private final ScoreSaberService scoreSaberService;

    /**
     * The user service to use
     */
    @NonNull
    private final UserService userService;


    @Autowired
    public ScoreSaberPlatform(@NonNull ScoreSaberService scoreSaberService, @NonNull UserService userService) {
        super(Platforms.SCORESABER, 1, Map.of(
                1, new CurvePoint[]{
                        new CurvePoint(0, 0),
                        new CurvePoint(0.6, 0.18223233667439062),
                        new CurvePoint(0.65, 0.5866010012767576),
                        new CurvePoint(0.7, 0.6125565959114954),
                        new CurvePoint(0.75, 0.6451808210101443),
                        new CurvePoint(0.8, 0.6872268862950283),
                        new CurvePoint(0.825, 0.7150465663454271),
                        new CurvePoint(0.85, 0.7462290664143185),
                        new CurvePoint(0.875, 0.7816934560296046),
                        new CurvePoint(0.9, 0.825756123560842),
                        new CurvePoint(0.91, 0.8488375988124467),
                        new CurvePoint(0.92, 0.8728710341448851),
                        new CurvePoint(0.93, 0.9039994071865736),
                        new CurvePoint(0.94, 0.9417362980580238),
                        new CurvePoint(0.95, 1),
                        new CurvePoint(0.955, 1.0388633331418984),
                        new CurvePoint(0.96, 1.0871883573850478),
                        new CurvePoint(0.965, 1.1552120359501035),
                        new CurvePoint(0.97, 1.2485807759957321),
                        new CurvePoint(0.9725, 1.3090333065057616),
                        new CurvePoint(0.975, 1.3807102743105126),
                        new CurvePoint(0.9775, 1.4664726399289512),
                        new CurvePoint(0.98, 1.5702410055532239),
                        new CurvePoint(0.9825, 1.697536248647543),
                        new CurvePoint(0.985, 1.8563887693647105),
                        new CurvePoint(0.9875, 2.058947159052738),
                        new CurvePoint(0.99, 2.324506282149922),
                        new CurvePoint(0.99125, 2.4902905794106913),
                        new CurvePoint(0.9925, 2.685667856592722),
                        new CurvePoint(0.99375, 2.9190155639254955),
                        new CurvePoint(0.995, 3.2022017597337955),
                        new CurvePoint(0.99625, 3.5526145337555373),
                        new CurvePoint(0.9975, 3.996793606763322),
                        new CurvePoint(0.99825, 4.325027383589547),
                        new CurvePoint(0.999, 4.715470646416203),
                        new CurvePoint(0.9995, 5.019543595874787),
                        new CurvePoint(1, 5.367394282890631),
                }
        ));
        this.scoreSaberService = scoreSaberService;
        this.userService = userService;
    }

    /**
     * Gets the modifier for the given accuracy.
     *
     * @param accuracy The accuracy.
     * @return The modifier.
     */
    public double getModifier(double accuracy) {
        accuracy = MathUtils.clamp(accuracy, 0, 100) / 100;
        CurvePoint[] curve = this.getCurve(this.getCurrentCurveVersion());
        if (accuracy <= 0) {
            return 0;
        }

        if (accuracy >= 1) {
            return curve[curve.length - 1].getMultiplier();
        }

        for (int i = 0; i < curve.length - 1; i++) {
            CurvePoint point = curve[i];
            CurvePoint nextPoint = curve[i + 1];
            if (accuracy >= point.getAcc() && accuracy <= nextPoint.getAcc()) {
                return MathUtils.lerp(
                        point.getMultiplier(), nextPoint.getMultiplier(),
                        (accuracy - point.getAcc()) / (nextPoint.getAcc() - point.getAcc())
                );
            }
        }

        return 0;
    }

    @Override
    public double getPp(double stars, double accuracy) {
        if (accuracy <= 1) { // Convert the accuracy to a percentage
            accuracy *= 100;
        }
        double pp = stars * this.starMultiplier;
        return this.getModifier(accuracy) * pp;
    }

    @Override
    public void updatePlayers() {
        for (User user : this.userService.getUsers(false)) {
            if (!user.isLinkedAccount()) { // Check if the user has linked their account
                continue;
            }
            ScoreSaberAccountToken account = scoreSaberService.getAccount(user); // Get the account from the ScoreSaber API
            user.addStatistic(this.getPlatform(), DateUtils.alignToCurrentHour(new Date()), new ScoreSaberStatistic(
                    account.getRank(),
                    account.getCountryRank(),
                    account.getScoreStats().getTotalScore(),
                    account.getScoreStats().getTotalRankedScore(),
                    account.getScoreStats().getAverageRankedAccuracy(),
                    account.getScoreStats().getTotalPlayCount()
            )); // Add the statistic to the user's history
            this.userService.saveUser(user); // Save the user
        }
    }

    @Override
    public void updateMetrics() {
        // todo: switch to InfluxDB?

//        try (Sender sender = questDBService.getSender()) {
//            TotalScoresResponse totalScores = ScoreService.INSTANCE.getTotalScores(this.getPlatform());
//            sender.table("metrics")
//                    .symbol("platform", this.getPlatform().getPlatformName())
//                    .longColumn("total_scores", totalScores.getTotalScores())
//                    .longColumn("total_ranked_scores", totalScores.getTotalRankedScores())
//                    .atNow();
//        }
    }

    @Override
    public void updateLeaderboards() {
        List<Score> scores = ScoreService.INSTANCE.getRankedScores(this.getPlatform());
        Map<String, ScoreSaberLeaderboardToken> leaderboards = new HashMap<>();
        for (ScoreSaberLeaderboardPageToken rankedLeaderboard : this.scoreSaberService.getRankedLeaderboards()) {
            for (ScoreSaberLeaderboardToken leaderboard : rankedLeaderboard.getLeaderboards()) {
                leaderboards.put(leaderboard.getId(), leaderboard);
            }
        }

        // Add any missing leaderboards
        for (Score score : scores) {
            if (leaderboards.containsKey(score.getLeaderboardId())) {
                continue;
            }

            ScoreSaberLeaderboardToken leaderboard = this.scoreSaberService.getLeaderboard(score.getLeaderboardId(), true);
            leaderboards.put(leaderboard.getId(), leaderboard);
        }

        log.info("Updating {} leaderboards for platform '{}'",
                leaderboards.size(),
                this.getPlatform().getPlatformName()
        );

        // Update the leaderboards
        int finished = 0;
        for (Map.Entry<String, ScoreSaberLeaderboardToken> leaderboardEntry : leaderboards.entrySet()) {
            String id = leaderboardEntry.getKey();
            ScoreSaberLeaderboardToken leaderboard = leaderboardEntry.getValue();
            try {
                List<Score> toUpdate = scores.stream().filter(score -> {
                    if (!score.getLeaderboardId().equals(id)) { // Check if the leaderboard ID matches
                        return false;
                    }
                    double pp = this.getPp(leaderboard.getStars(), score.getAccuracy());
                    return pp != score.getPp(); // Check if the pp has changed
                }).toList();

                for (Score score : toUpdate) { // Update the scores
                    if (leaderboard.getStars() == 0) { // The leaderboard was unranked
                        score.setPp(0D);
                    }
                    double pp = this.getPp(leaderboard.getStars(), score.getAccuracy());
                    score.setPp(pp);
                }

                if (!toUpdate.isEmpty()) { // Save the scores
                    ScoreService.INSTANCE.updateScores(toUpdate.toArray(new Score[0]));
                }
                finished++;

                if (finished % 100 == 0 || finished == leaderboards.size()) {
                    log.info("Updated {}/{} leaderboards for platform '{}'",
                            finished,
                            leaderboards.size(),
                            this.getPlatform().getPlatformName()
                    );
                }
            } catch (Exception ex) {
                log.error("An error occurred while updating leaderboard '{}'", id, ex);
            }
        }
    }
}
