package cc.fascinated.repository;

import cc.fascinated.model.score.Score;
import cc.fascinated.platform.Platform;
import lombok.NonNull;
import org.springframework.data.mongodb.repository.Aggregation;
import org.springframework.data.mongodb.repository.MongoRepository;
import org.springframework.data.repository.query.Param;

import java.util.List;

/**
 * @author Fascinated (fascinated7)
 */
public interface ScoreRepository extends MongoRepository<Score, String> {

    /**
     * Gets the top ranked scores from the platform.
     *
     * @param platform the platform to get the scores from
     * @param amount   the amount of scores to get
     * @return the scores
     */
    @Aggregation(pipeline = {
            "{ $match: { platform: ?0, pp: { $gt: 0 } } }",
            "{ $sort: { pp: -1 } }",
            "{ $limit: ?1 }"
    })
    List<Score> getTopRankedScores(@NonNull Platform.Platforms platform, int amount);

    /**
     * Gets all the ranked scores from the platform.
     *
     * @param platform the platform to get the scores from
     * @return the scores
     */
    @Aggregation(pipeline = {
            "{ $match: { platform: ?0, pp: { $gt: 0 } } }",
            "{ $sort: { pp: -1 } }"
    })
    List<Score> getRankedScores(@NonNull Platform.Platforms platform);

    /**
     * Updates a scores pp value.
     *
     * @param id The id of the score to update
     * @param pp The new pp value
     */
    @Aggregation(pipeline = {
            "{ $match: { _id: ?0 } }",
            "{ $set: { pp: ?1 } }"
    })
    void updateScorePP(@Param("id") long id, @Param("pp") double pp);

    /**
     * Gets the total scores for the platform.
     *
     * @param platform the platform to get the scores from
     * @return the total scores
     */
    @Aggregation(pipeline = {
            "{ $match: { platform: ?0 } }",
            "{ $count: 'total' }"
    })
    int getTotalScores(@NonNull Platform.Platforms platform);

    /**
     * Gets the total ranked scores for the platform.
     *
     * @param platform the platform to get the scores from
     * @return the total ranked scores
     */
    @Aggregation(pipeline = {
            "{ $match: { platform: ?0, pp: { $gt: 0 } } }",
            "{ $count: 'total' }"
    })
    int getTotalRankedScores(@NonNull Platform.Platforms platform);
}
