package cc.fascinated.services;

import cc.fascinated.common.DateUtils;
import cc.fascinated.common.EnumUtils;
import cc.fascinated.common.MathUtils;
import cc.fascinated.model.leaderboard.Leaderboard;
import cc.fascinated.model.score.Score;
import cc.fascinated.model.score.TotalScoresResponse;
import cc.fascinated.model.score.impl.scoresaber.ScoreSaberScore;
import cc.fascinated.model.score.impl.scoresaber.ScoreSaberScoreResponse;
import cc.fascinated.model.token.ScoreSaberLeaderboardToken;
import cc.fascinated.model.token.ScoreSaberPlayerScoreToken;
import cc.fascinated.model.token.ScoreSaberScoreToken;
import cc.fascinated.model.user.User;
import cc.fascinated.model.user.UserDTO;
import cc.fascinated.platform.Platform;
import cc.fascinated.repository.ScoreRepository;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Fascinated (fascinated7)
 */
@Service
@Log4j2(topic = "Score Service")
public class ScoreService {
    public static ScoreService INSTANCE;

    /**
     * The counter service to use.
     */
    @NonNull
    private final CounterService counterService;

    /**
     * The user service to use.
     */
    @NonNull
    private final UserService userService;

    /**
     * The ScoreSaber service to use.
     */
    @NonNull
    private final ScoreSaberService scoreSaberService;

    /**
     * The score repository to use.
     */
    @NonNull
    private final ScoreRepository scoreRepository;

    @Autowired
    public ScoreService(@NonNull CounterService counterService, @NonNull UserService userService, @NonNull ScoreSaberService scoreSaberService,
                        @NonNull ScoreRepository scoreRepository) {
        INSTANCE = this;
        this.counterService = counterService;
        this.userService = userService;
        this.scoreSaberService = scoreSaberService;
        this.scoreRepository = scoreRepository;
    }

    /**
     * Gets the top ranked scores from the platform.
     *
     * @param platform   The platform to get the scores from.
     * @param amount     The amount of scores to get.
     * @param scoresOnly Whether to only get the scores.
     * @return The scores.
     */
    public List<ScoreSaberScoreResponse> getTopRankedScores(@NonNull Platform.Platforms platform, int amount, boolean scoresOnly) {
        List<Score> trackedScores = scoreRepository.getTopRankedScores(platform, amount);
        List<ScoreSaberScoreResponse> scores = new ArrayList<>();
        for (Score score : trackedScores) {
            ScoreSaberScore scoreSaberScore = (ScoreSaberScore) score;
            UserDTO user = scoresOnly ? null : userService.getUser(score.getPlayerId()).getAsDTO();
            Leaderboard leaderboard = scoresOnly ? null : Leaderboard.getFromScoreSaberToken(scoreSaberService.getLeaderboard(score.getLeaderboardId()));

            scores.add(new ScoreSaberScoreResponse(
                    score.getId(),
                    score.getPlayerId(),
                    score.getPlatform(),
                    score.getPlatformScoreId(),
                    score.getLeaderboardId(),
                    score.getRank(),
                    score.getAccuracy(),
                    score.getPp(),
                    score.getScore(),
                    score.getModifiers(),
                    score.getMisses(),
                    score.getBadCuts(),
                    score.getTimestamp(),
                    scoreSaberScore.getWeight(),
                    scoreSaberScore.getMultiplier(),
                    scoreSaberScore.getMaxCombo(),
                    user,
                    leaderboard
            ));
        }

        return scores;
    }

    /**
     * Gets all the ranked scores from the platform.
     *
     * @param platform The platform to get the scores from.
     * @return The scores.
     */
    public List<Score> getRankedScores(@NonNull Platform.Platforms platform) {
        return scoreRepository.getRankedScores(platform);
    }

    /**
     * Updates a scores pp value.
     *
     * @param scores The scores to update.
     */
    public void updateScores(Score... scores) {
        for (Score score : scores) {
            scoreRepository.updateScorePP(score.getId(), score.getPp());
        }
    }

    public TotalScoresResponse getTotalScores(Platform.Platforms platform) {
        return new TotalScoresResponse(
                scoreRepository.getTotalScores(platform),
                scoreRepository.getTotalRankedScores(platform)
        );
    }

    /**
     * Tracks a ScoreSaber score.
     *
     * @param token The token of the score to track.
     */
    public void trackScoreSaberScore(ScoreSaberPlayerScoreToken token) {
        ScoreSaberLeaderboardToken leaderboard = this.scoreSaberService.getLeaderboard(token.getLeaderboard().getId());
        ScoreSaberScoreToken score = token.getScore();
        User user = userService.getUser(score.getLeaderboardPlayerInfo().getId());

        double accuracy = leaderboard.getMaxScore() != 0 ? ((double) score.getBaseScore() / leaderboard.getMaxScore()) * 100 : 0;
        if (accuracy == 0) {
            log.warn("[Scoresaber] Leaderboard '{}' has a max score of 0, unable to calculate accuracy :(", leaderboard.getId());
        }
        double pp = score.getPp() != 0 ? PlatformService.INSTANCE.getScoreSaberPlatform().getPp(leaderboard.getStars(), accuracy) : 0; // Recalculate the pp
        String[] modifiers = !score.getModifiers().isEmpty() ? score.getModifiers().split(",") : new String[0];

        ScoreSaberScore scoreSaberScore = new ScoreSaberScore(
                counterService.getNext(CounterService.CounterType.SCORE),
                user.getSteamId(),
                Platform.Platforms.SCORESABER,
                score.getId(),
                leaderboard.getId(),
                score.getRank(),
                accuracy,
                pp,
                score.getBaseScore(),
                modifiers,
                score.getMissedNotes(),
                score.getBadCuts(),
                DateUtils.getDateFromString(score.getTimeSet()),
                score.getWeight(),
                score.getMultiplier(),
                score.getMaxCombo()
        );
        scoreRepository.save(scoreSaberScore);
        this.logScore(Platform.Platforms.SCORESABER, scoreSaberScore, user);
    }

    /**
     * Logs a score.
     *
     * @param platform The platform the score was tracked on.
     * @param score    The score to log.
     * @param user     The user who set the score.
     */
    private void logScore(@NonNull Platform.Platforms platform, @NonNull Score score, @NonNull User user) {
        String platformName = EnumUtils.getEnumName(platform);
        boolean isRanked = score.getPp() != 0;
        log.info("[{}] Tracked{} Score! id: {}, acc: {}%, {} score id: {},{} leaderboard: {}, player: {}",
                platformName,
                isRanked ? " Ranked" : "",
                score.getId(),
                MathUtils.format(score.getAccuracy(), 2),
                platformName.toLowerCase(), score.getPlatformScoreId(),
                isRanked ? " pp: %s,".formatted(score.getPp()) : "",
                score.getLeaderboardId(),
                user.getUsername() == null ? user.getSteamId() : user.getUsername()
        );
    }
}
