package cc.fascinated.services;

import cc.fascinated.common.TimeUtils;
import cc.fascinated.exception.impl.BadRequestException;
import cc.fascinated.model.token.ScoreSaberAccountToken;
import cc.fascinated.model.user.ScoreSaberAccount;
import cc.fascinated.model.user.User;
import cc.fascinated.repository.UserRepository;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

/**
 * @author Fascinated (fascinated7)
 */
@Service
@Log4j2(topic = "User Service")
public class UserService {
    /**
     * The interval to force update the user's account.
     */
    private static long FORCE_UPDATE_INTERVAL = TimeUnit.HOURS.toMillis(4);

    /**
     * The user repository to use
     */
    @NonNull
    private final UserRepository userRepository;

    /**
     * The ScoreSaber service to use
     */
    @NonNull
    private final ScoreSaberService scoreSaberService;

    @Autowired
    public UserService(@NonNull UserRepository userRepository, @NonNull ScoreSaberService scoreSaberService) {
        this.userRepository = userRepository;
        this.scoreSaberService = scoreSaberService;
    }

    /**
     * Gets a user by their id
     *
     * @param steamId the id of the user's steam profile
     * @return the user
     * @throws BadRequestException if the user is not found
     */
    public User getUser(String steamId) {
        if (!this.isValidSteamId(steamId)) {
            throw new BadRequestException("Invalid steam id");
        }
        Optional<User> userOptional = this.userRepository.findBySteamId(steamId);
        User user;
        boolean shouldUpdate = false;
        if (userOptional.isEmpty()) {
            // todo: check the steam API to see if the user exists
            user = new User(UUID.randomUUID());
            user.setSteamId(steamId);
            shouldUpdate = true;
        } else {
            user = userOptional.get();
        }

        // Ensure the users ScoreSaber account is up-to-date
        ScoreSaberAccount scoresaberAccount = user.getScoresaberAccount();
        if (scoresaberAccount == null || scoresaberAccount.getLastUpdated().before(new Date(System.currentTimeMillis() - FORCE_UPDATE_INTERVAL))) {
            log.info("Updating account for '{}', last update: {}",
                    steamId,
                    scoresaberAccount == null ? "now" : TimeUtils.format(System.currentTimeMillis() - scoresaberAccount.getLastUpdated().getTime())
            );
            ScoreSaberAccountToken accountToken = scoreSaberService.getAccount(user);
            user.setScoresaberAccount(ScoreSaberAccount.getFromToken(accountToken)); // Update the ScoreSaber account
            user.setUsername(accountToken.getName()); // Update the username
            shouldUpdate = true;
        }
        if (shouldUpdate) {
            this.saveUser(user); // Save the user
        }
        return user;
    }

    /**
     * Saves a user to the database
     *
     * @param user the user to save
     */
    public void saveUser(User user) {
        this.userRepository.save(user);
    }

    /**
     * Gets all users in the database
     *
     * @return all users
     */
    public List<User> getUsers(boolean steamIdsOnly) {
        if (steamIdsOnly) {
            return this.userRepository.fetchOnlySteamIds();
        }
        return this.userRepository.findAll();
    }

    /**
     * Validates a steam id
     *
     * @param steamId the steam id to validate
     * @return if the steam id is valid
     */
    public boolean isValidSteamId(String steamId) {
        return steamId != null && steamId.length() == 17;
    }
}
